from happytransformer import HappyTextToText, TTSettings, TTTrainArgs, TTEvalArgs



def test_default_simple():
    happy_tt = HappyTextToText()
    output = happy_tt.generate_text("translate English to French: Hello my name is Eric")
    assert type(output.text) == str

def test_default_min_max_length():
    happy_tt = HappyTextToText()
    args = TTSettings(min_length=5, max_length=5)
    output = happy_tt.generate_text("translate English to French: Hello my name is Eric", args=args)
    tokens = happy_tt.tokenizer.encode(output.text, return_tensors="pt")
    length = len(tokens[0])
    assert length == 5

def test_all_methods():
    happy_tt = HappyTextToText()

    greedy_settings = TTSettings(min_length=5, max_length=5, no_repeat_ngram_size=2)
    output_greedy = happy_tt.generate_text(
        "translate English to French: Hello my name is Eric",
        args=greedy_settings)


    beam_settings = TTSettings(min_length=5, max_length=5, early_stopping=True, num_beams=5)

    output_beam_search = happy_tt.generate_text(
        "translate English to French: Hello my name is Eric",
        args=beam_settings)

    generic_sampling_settings = TTSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=0, temperature=0.7)

    output_generic_sampling = happy_tt.generate_text(
        "translate English to French: Hello my name is Eric",
        args=generic_sampling_settings)

    top_k_sampling_settings = TTSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=50, temperature=0.7)

    output_top_k_sampling = happy_tt.generate_text(
        "translate English to French: Hello my name is Eric",
        args=top_k_sampling_settings)

    top_p_sampling_settings = TTSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=0, top_p=0.8, temperature=0.7)

    output_top_p_sampling = happy_tt.generate_text(
        "translate English to French: Hello my name is Eric",
        args=top_p_sampling_settings)

    assert type(output_greedy.text) == str
    assert type(output_beam_search.text) == str
    assert type(output_generic_sampling.text) == str
    assert type(output_top_k_sampling.text) == str
    assert type(output_top_p_sampling.text) == str

    print("greedy: ", output_greedy.text, end="\n\n")
    print("beam-search: ", output_beam_search.text, end="\n\n")
    print("generic-sampling: ", output_generic_sampling.text, end="\n\n")
    print("top-k-sampling: ", output_top_k_sampling.text, end="\n\n")
    print("top-p-sampling: ", output_top_p_sampling.text, end="\n\n")

def test_tt_save():
    happy_tt = HappyTextToText()
    happy_tt.save("model/")
    result_before = happy_tt.generate_text("translate English to French: Hello my name is Eric")

    happy = HappyTextToText(load_path="model/")
    result_after = happy.generate_text("translate English to French: Hello my name is Eric")

    assert result_before.text == result_after.text


def test_tt_train_simple():
    happy_tt = HappyTextToText()
    happy_tt.train("../data/tt/train-eval-grammar.csv")

def test_tt_eval_simple():
    happy_tt = HappyTextToText()
    happy_tt.eval("../data/tt/train-eval-grammar.csv")

def test_tt_eval_loss_decreases():
    happy_tt = HappyTextToText()

    before_result = happy_tt.eval("../data/tt/train-eval-grammar.csv")
    happy_tt.train("../data/tt/train-eval-grammar.csv")
    after_result = happy_tt.eval("../data/tt/train-eval-grammar.csv")

    assert before_result.loss > after_result.loss

def test_tt_train_custom_p():
    happy_tt = HappyTextToText()
    args = TTTrainArgs(num_train_epochs=2, max_input_length=100, max_output_length=100, preprocessing_processes=4, batch_size=2)
    happy_tt.train("../data/tt/train-eval-grammar.csv", args=args)

def test_tt_eval_custom_p():
    happy_tt = HappyTextToText()
    args = TTEvalArgs(max_input_length=100, max_output_length=100, preprocessing_processes=4, batch_size=2)
    result = happy_tt.eval("../data/tt/train-eval-grammar.csv", args=args)
    assert type(result.loss) is float


from typing import List, Optional
from dataclasses import dataclass

from transformers import TokenClassificationPipeline, AutoModelForTokenClassification

from happytransformer.happy_transformer import HappyTransformer
from happytransformer.toc.trainer import TOCTrainer
from happytransformer.cuda_detect import detect_cuda_device_number
from happytransformer.adaptors import get_adaptor
from happytransformer.toc import  ARGS_TOC_TRAIN, ARGS_TOC_EVAl, ARGS_TOC_TEST


@dataclass
class TokenClassificationResult:
    word: str
    score: float
    entity: str
    index: int
    start: int
    end: int


class HappyTokenClassification(HappyTransformer):
    """
    A user facing class for token classification
    """
    def __init__(
        self, model_type: str = "BERT", model_name: str = "dslim/bert-base-NER", load_path: str = "", use_auth_token: str = None):

        self.adaptor = get_adaptor(model_type)

        if load_path != "":
            model = AutoModelForTokenClassification.from_pretrained(load_path)
        else:
            model = AutoModelForTokenClassification.from_pretrained(model_name, use_auth_token=use_auth_token)


        super().__init__(model_type, model_name, model)

        device_number = detect_cuda_device_number()

        self._pipeline = TokenClassificationPipeline(model=self.model, tokenizer=self.tokenizer, device=device_number)

        self._trainer = TOCTrainer(self.model, model_type, self.tokenizer, self._device, self.logger)

    def classify_token(self, text: str) -> List[TokenClassificationResult]:
        """
        :param text: Text that contains tokens to be classified
        :return:
        """
        if not isinstance(text, str):
            raise ValueError('the "text" argument must be a single string')
        results = self._pipeline(text)

        return [
            TokenClassificationResult(
                word=answer["word"],
                score=answer["score"],
                entity=answer["entity"],
                index=answer["index"],
                start=answer["start"],
                end=answer["end"],
            )
            for answer in results
        ]


    def train(self, input_filepath, args=ARGS_TOC_TRAIN):
        raise NotImplementedError("train() is currently not available")

    def eval(self, input_filepath, args=ARGS_TOC_EVAl):
        raise NotImplementedError("eval() is currently not available")

    def test(self, input_filepath, args=ARGS_TOC_TEST):
        raise NotImplementedError("test() is currently not available")



from happytransformer import HappyTokenClassification
from happytransformer.happy_token_classification import TokenClassificationResult


def test_classify_text():
    happy_toc = HappyTokenClassification(model_type="BERT", model_name="dslim/bert-base-NER")
    expected_result = [TokenClassificationResult(word='Geoffrey', score=0.9988969564437866, entity='B-PER', index=4, start=11, end=19), TokenClassificationResult(word='Toronto', score=0.9993201494216919, entity='B-LOC', index=9, start=34, end=41)]

    result = happy_toc.classify_token("My name is Geoffrey and I live in Toronto")
    assert result == expected_result

def test_toc_save():
    happy = HappyTokenClassification()
    happy.save("model/")
    result_before = happy.classify_token("My name is Geoffrey and I live in Toronto")

    happy = HappyTokenClassification(load_path="model/")
    result_after = happy.classify_token("My name is Geoffrey and I live in Toronto")

    assert result_before[0].word==result_after[0].word



from happytransformer.happy_question_answering import HappyQuestionAnswering
from happytransformer.happy_word_prediction import HappyWordPrediction
from happytransformer.happy_text_classification import HappyTextClassification
from happytransformer.happy_next_sentence import HappyNextSentence
from happytransformer.happy_token_classification import HappyTokenClassification
from happytransformer.happy_generation import HappyGeneration, GENSettings
from happytransformer.happy_text_to_text import HappyTextToText, TTSettings

from happytransformer.gen.default_args import ARGS_GEN_TRAIN, ARGS_GEN_EVAl
from happytransformer.qa.default_args import ARGS_QA_TRAIN, ARGS_QA_EVAl, ARGS_QA_TEST
from happytransformer.tc.default_args import ARGS_TC_TRAIN, ARGS_TC_EVAL, ARGS_TC_TEST
from happytransformer.wp.default_args import ARGS_WP_TRAIN, ARGS_WP_EVAl

from happytransformer.gen.trainer import GENTrainArgs, GENEvalArgs
from happytransformer.qa.trainer import QATestArgs, QAEvalArgs, QATrainArgs
from happytransformer.tc.trainer import TCTrainArgs, TCEvalArgs, TCTestArgs
from happytransformer.wp.trainer import WPTrainArgs, WPEvalArgs
from happytransformer.tt.trainer import TTTrainArgs, TTEvalArgs


from happytransformer.happy_generation import (
    HappyGeneration, GENSettings
)

name = "happytransformer"



from happytransformer import (
    HappyGeneration,
    ARGS_GEN_TRAIN,
    ARGS_GEN_EVAl,
    GENSettings,
    GENTrainArgs,
    GENEvalArgs
)

from tests.shared_tests import run_save_load

def test_default_simple():
    happy_gen = HappyGeneration()
    args = GENSettings(min_length=5, max_length=5)
    output = happy_gen.generate_text("Artificial intelligence is ", args=args)
    assert type(output.text) == str


def test_default_min_max_length():
    happy_gen = HappyGeneration()
    args = GENSettings(min_length=5, max_length=5)
    output = happy_gen.generate_text("Artificial intelligence is ", args=args)
    tokens = happy_gen.tokenizer.encode(output.text, return_tensors="pt")
    length = len(tokens[0])
    assert length == 5

def test_bad_words():
    happy_gen = HappyGeneration()
    # Test single words
    args_test_word_single = GENSettings(bad_words=["new", "tool"])
    # result without bad_words:  "a new field of research that has been gaining momentum"
    output_single = happy_gen.generate_text("Artificial intelligence is ", args=args_test_word_single)
    output_words_single = output_single.text.split()
    for phrase in args_test_word_single.bad_words:
        for word in phrase.split():
            assert word not in output_words_single
    # Test phrases 
    args_test_word_phrase = GENSettings(bad_words=["new field"])
    output_phrase = happy_gen.generate_text("Artificial intelligence is ", args=args_test_word_phrase)
    for phrase in args_test_word_phrase.bad_words:
        assert phrase not in output_phrase.text


def test_top_p():
    happy_gen = HappyGeneration()
    # Test small values
    args_small = GENSettings(top_p=0.01,  max_length=5)
    output_small = happy_gen.generate_text("Artificial intelligence is ", args=args_small)
    tokens_small = happy_gen.tokenizer.encode(output_small.text, return_tensors="pt")
    length_small = len(tokens_small[0])
    assert length_small == 5

    # Test large values
    args_large = GENSettings(top_p=1, max_length=5)
    output_large = happy_gen.generate_text("Artificial intelligence is ", args=args_large)
    tokens_large = happy_gen.tokenizer.encode(output_large.text, return_tensors="pt")
    length_large = len(tokens_large[0])
    assert length_large == 5



def test_all_methods():
    happy_gen = HappyGeneration()

    greedy_settings = GENSettings(min_length=5, max_length=5, no_repeat_ngram_size=2)
    output_greedy = happy_gen.generate_text(
        "Artificial intelligence is ",
        args=greedy_settings)


    beam_settings = GENSettings(min_length=5, max_length=5, early_stopping=True, num_beams=5)

    output_beam_search = happy_gen.generate_text(
        "Artificial intelligence is ",
        args=beam_settings)

    generic_sampling_settings = GENSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=0, temperature=0.7)

    output_generic_sampling = happy_gen.generate_text(
        "Artificial intelligence is ",
        args=generic_sampling_settings)

    top_k_sampling_settings = GENSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=50, temperature=0.7)

    output_top_k_sampling = happy_gen.generate_text(
        "Artificial intelligence is ",
        args=top_k_sampling_settings)

    top_p_sampling_settings = GENSettings(min_length=5, max_length=5, do_sample=True, early_stopping=False, top_k=0, top_p=0.8, temperature=0.7)

    output_top_p_sampling = happy_gen.generate_text(
        "Artificial intelligence is ",
        args=top_p_sampling_settings)

    assert type(output_greedy.text) == str
    assert type(output_beam_search.text) == str
    assert type(output_generic_sampling.text) == str
    assert type(output_top_k_sampling.text) == str
    assert type(output_top_p_sampling.text) == str

    print("greedy: ", output_greedy.text, end="\n\n")
    print("beam-search: ", output_beam_search.text, end="\n\n")
    print("generic-sampling: ", output_generic_sampling.text, end="\n\n")
    print("top-k-sampling: ", output_top_k_sampling.text, end="\n\n")
    print("top-p-sampling: ", output_top_p_sampling.text, end="\n\n")

def test_gen_train_basic():
    happy_gen = HappyGeneration()
    happy_gen.train("../data/gen/train-eval.txt")

def test_gen_eval_basic():
    happy_gen = HappyGeneration()
    result = happy_gen.eval("../data/gen/train-eval.txt")
    assert type(result.loss) == float

def test_gen_train_effectiveness_multi():
    happy_gen = HappyGeneration()
    before_result = happy_gen.eval("../data/gen/train-eval.txt")
    happy_gen.train("../data/gen/train-eval.txt")
    after_result = happy_gen.eval("../data/gen/train-eval.txt")

    assert after_result.loss < before_result.loss

def test_gen_save_load_train():
    happy_gen = HappyGeneration()
    output_path = "data/gen-train.txt"
    data_path = "../data/gen/train-eval.txt"
    run_save_load(happy_gen, output_path, ARGS_GEN_TRAIN, data_path, "train")

def test_gen_save_load_eval():
    happy_gen = HappyGeneration()
    output_path = "data/wp-eval.txt"
    data_path = "../data/gen/train-eval.txt"
    run_save_load(happy_gen, output_path, ARGS_GEN_EVAl, data_path, "eval")

def test_gen_save():
    happy = HappyGeneration()
    happy.save("model/")
    result_before = happy.generate_text("Natural language processing is")

    happy = HappyGeneration(load_path="model/")
    result_after=happy.generate_text("Natural language processing is")

    assert result_before == result_after

def test_wp_train_eval_with_dic():

    happy_gen = HappyGeneration()
    train_args = {'learning_rate': 0.01,  "num_train_epochs": 1}


    happy_gen.train("../data/gen/train-eval.txt" , args=train_args)
    eval_args = {}

    result = happy_gen.eval("../data/gen/train-eval.txt", args=eval_args)
    assert type(result.loss) == float


def test_gen_train_eval_with_dataclass():

    happy_gen = HappyGeneration()
    train_args = GENTrainArgs(learning_rate=0.01, num_train_epochs=1)

    happy_gen.train("../data/gen/train-eval.txt" , args=train_args)

    eval_args = GENEvalArgs()

    after_result = happy_gen.eval("../data/wp/train-eval.txt", args=eval_args)

    assert type(after_result.loss) == float

def test_generate_after_train_eval():
    happy_gen = HappyGeneration()
    happy_gen.train("../data/gen/train-eval.txt")
    eval_result = happy_gen.eval("../data/gen/train-eval.txt")
    output = happy_gen.generate_text("Artificial intelligence is ")
    assert type(output.text) == str




"""
Contains the HappyQuestionAnswering class.

"""
from typing import List
from dataclasses import dataclass
from transformers import QuestionAnsweringPipeline, AutoModelForQuestionAnswering

from happytransformer.happy_transformer import HappyTransformer
from happytransformer.qa.trainer import QATrainer, QATrainArgs, QAEvalArgs, QATestArgs
from happytransformer.happy_trainer import EvalResult
from happytransformer.qa import ARGS_QA_TRAIN, ARGS_QA_EVAl, ARGS_QA_TEST

from happytransformer.cuda_detect import detect_cuda_device_number
from happytransformer.adaptors import get_adaptor
from happytransformer.fine_tuning_util import create_args_dataclass

@dataclass
class QuestionAnsweringResult:
    answer: str
    score: float
    start: int
    end: int


class HappyQuestionAnswering(HappyTransformer):
    """
    This class is a user facing class that allows users to solve question answering problems using
    a transformer QA models. These models are able to answer a question given context for the
    question by selecting a span within the context that answers the question.

    The purpose of this class is to be lightweight and easy
    to understand and to offload complex tasks to
    other classes.
    """
    def __init__(self, model_type="DISTILBERT",
                 model_name="distilbert-base-cased-distilled-squad", 
                 load_path: str = "",
                 use_auth_token: str = None):
        
        self.adaptor = get_adaptor(model_type)

        if load_path != "":
            model = AutoModelForQuestionAnswering.from_pretrained(load_path)
        else:
            model = AutoModelForQuestionAnswering.from_pretrained(model_name, use_auth_token=use_auth_token)

        super().__init__(model_type, model_name, model)
        device_number = detect_cuda_device_number()

        self._pipeline = QuestionAnsweringPipeline(model=self.model, tokenizer=self.tokenizer, device=device_number)

        self._trainer = QATrainer(self.model, model_type, self.tokenizer, self._device, self.logger)

    def answer_question(self, context: str, question: str, top_k: int = 1) \
            -> List[QuestionAnsweringResult]:
        """
        Find the answers to a question.
        The answer MUST be contained somewhere within the context for this to work.
        top_k describes the number of answers to return.
        """

        pipeline_output = self._pipeline(context=context, question=question, topk=top_k)
        # transformers returns a single dictionary when top_k ==1.
        # Our convention however is to have constant output format
        answers = [pipeline_output] if top_k == 1 else pipeline_output

        return [
            QuestionAnsweringResult(
                answer=answer["answer"],
                score=answer["score"],
                start=answer["start"],
                end=answer["end"],)
            for answer in answers
        ]

    def train(self, input_filepath, args=QATrainArgs()):
        """
        Trains the question answering model

        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values: context,
        question, answer_text, answer_start

        args: Either a QATrainArgs() object or a dictionary that contains all of the same keys as ARGS_QA_TRAIN

        return: None
        """

        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_QA_TRAIN,
                                                                input_dic_args=args,
                                                                method_dataclass_args=QATrainArgs)
        elif type(args) == QATrainArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a QATrainArgs object or a dictionary")


        self._trainer.train(input_filepath=input_filepath, dataclass_args=method_dataclass_args)

    def eval(self, input_filepath, args=QAEvalArgs()) -> EvalResult:
        """
        Trains the question answering model

        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values:
        context, question, answer_text, answer_start

        args: Either a QAEvalArgs() object or a dictionary that contains all of the same keys as ARGS_QA_EVAl

        return: A dictionary that contains a key called "eval_loss"

        """
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_QA_EVAl,
                                                                input_dic_args=args,
                                                                method_dataclass_args=QAEvalArgs)
        elif type(args) == QAEvalArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a QAEvalArgs object or a dictionary")

        return self._trainer.eval(input_filepath=input_filepath, dataclass_args=method_dataclass_args)


    def test(self, input_filepath, args=QATestArgs()):
        """
        Tests the question answering model. Used to obtain results

        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values:
        context, question

        args: Either a QATestArgs() object or a dictionary that contains all of the same keys as ARGS_QA_TEST

        return: A list of dictionaries. Each dictionary
        contains the keys: "score", "start", "end" and "answer"
        """
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_QA_TEST,
                                                                input_dic_args=args,
                                                                method_dataclass_args=QATestArgs)
        elif type(args) == QATestArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a QATestArgs object or a dictionary")


        return self._trainer.test(input_filepath=input_filepath, solve=self.answer_question, dataclass_args=method_dataclass_args)



"""
Fine-tune Text-to-text models

Based on the following sources:
1. https://github.com/huggingface/notebooks/blob/master/examples/summarization.ipynb
2. https://github.com/huggingface/transformers/blob/master/examples/pytorch/summarization/run_summarization.py
3. https://github.com/huggingface/transformers/blob/master/examples/pytorch/translation/run_translation.py
4. https://huggingface.co/transformers/
5. https://huggingface.co/docs/datasets/
"""

from dataclasses import dataclass
from happytransformer.happy_trainer import HappyTrainer, EvalResult
from datasets import load_dataset
from transformers import DataCollatorForSeq2Seq, Seq2SeqTrainingArguments, Seq2SeqTrainer
import tempfile


@dataclass
class TTTrainArgs:
    """
    Used to adjust the settings when calling HappyTextToText.train()
    """
    learning_rate: float = 5e-5
    num_train_epochs: int = 3
    weight_decay: float = 0
    adam_beta1: float = 0.9
    adam_beta2: float = 0.999
    adam_epsilon: float = 1e-8
    max_grad_norm: float = 1.0
    save_preprocessed_data: bool = False
    save_preprocessed_data_path: str = ""
    load_preprocessed_data: bool = False
    load_preprocessed_data_path: str = ""
    preprocessing_processes: int = 1
    batch_size: int = 1
    max_input_length: int = 1024
    max_output_length: int = 1024


@dataclass
class TTEvalArgs:
    """
    Used to adjust the settings when calling HappyTextToText.eval()
    """
    batch_size: int = 1
    save_preprocessed_data: bool = False
    save_preprocessed_data_path: str = ""
    load_preprocessed_data: bool = ""
    load_preprocessed_data_path: str = ""
    preprocessing_processes: int = 1
    max_input_length: int = 1024
    max_output_length: int = 1024


@dataclass
class TTTestArgs:
    """
    Used to adjust the settings when calling HappyTextToText.test()
    """
    save_preprocessed_data: bool = False
    save_preprocessed_data_path: str = ""
    load_preprocessed_data: bool = False
    load_preprocessed_data_path: str = ""

class TTTrainer(HappyTrainer):
    """
    Trainer class for HappyTextToText
    """
    def __init__(self, model, model_type, tokenizer, device, logger):
        super().__init__(model, model_type, tokenizer, device, logger)
        self.__max_input_length = 1024
        self.__max_output_length = 1024


    def train(self, input_filepath, dataclass_args=TTTrainArgs):
        """
        :param input_filepath: A file path to a csv file that contains two columns: text_1 and text_2
        :param dataclass_args: A TTTrainArgs() object
        :return: None
        """

        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving preprocessed data is currently "
                             "not available for text-to-text models. "
                             )
        if dataclass_args.load_preprocessed_data:
            self.logger.info("Loading preprocessed data is currently "
                             "not available for text-to-text models. "
                             )

        self.logger.info("Preprocessing training data...")

        dataset = load_dataset("csv", data_files={"train": input_filepath}, delimiter=",")

        self.__max_input_length = dataclass_args.max_input_length
        self.__max_output_length = dataclass_args.max_output_length

        tokenized_dataset = dataset.map(
            self.__preprocess_function,
            batched=True,
            num_proc=dataclass_args.preprocessing_processes,
            remove_columns=["input", "target"],
        )

        self.logger.info("Training...")

        # Pads inputs and labels to max length
        data_collator = DataCollatorForSeq2Seq(self.tokenizer, model=self.model)

        # A temp dir is used so any files that are generated are deleted after training
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            training_args = Seq2SeqTrainingArguments(
                tmp_dir_name,
                do_train=True,
                do_eval=False,
                learning_rate=dataclass_args.learning_rate,
                weight_decay=dataclass_args.weight_decay,
                adam_beta1=dataclass_args.adam_beta1,
                adam_beta2=dataclass_args.adam_beta2,
                adam_epsilon=dataclass_args.adam_epsilon,
                max_grad_norm=dataclass_args.max_grad_norm,
                num_train_epochs=dataclass_args.num_train_epochs,
                report_to=["none"],
                save_strategy="no", # no checkpoints are saved
                per_device_train_batch_size=dataclass_args.batch_size
            )

            trainer = Seq2SeqTrainer(
                model=self.model,
                args=training_args,
                train_dataset=tokenized_dataset['train'],
                tokenizer=self.tokenizer,
                data_collator=data_collator,
            )
            trainer.train()


    def eval(self, input_filepath, dataclass_args=TTEvalArgs):
        """
        Evaluates the model by determining the loss.
        :param input_filepath: A file path to a csv file that contains two columns: text_1 and text_2
        :param dataclass_args: A TTEvalArgs() object
        :return: a EvalResult object that contains the loss
        """
        self.logger.info("Preprocessing evaluating data...")
        dataset = load_dataset("csv", data_files={"eval": input_filepath}, delimiter=",")

        self.__max_input_length = dataclass_args.max_input_length
        self.__max_output_length = dataclass_args.max_output_length

        tokenized_dataset = dataset.map(
            self.__preprocess_function,
            batched=True,
            remove_columns=["input", "target"],
            num_proc=dataclass_args.preprocessing_processes,
        )

        # Pads inputs and labels to max length
        data_collator = DataCollatorForSeq2Seq(self.tokenizer, model=self.model)

        # A temp dir is used so any files that are generated are deleted after evaluating
        with tempfile.TemporaryDirectory() as tmp_dir_name:
            eval_args = Seq2SeqTrainingArguments(
                tmp_dir_name,
                do_train=False,
                do_eval=True,
                seed=42,
                report_to=["none"],
                per_device_eval_batch_size=dataclass_args.batch_size,
            )

            trainer = Seq2SeqTrainer(
                model=self.model,
                args=eval_args,
                eval_dataset=tokenized_dataset['eval'],
                tokenizer=self.tokenizer,
                data_collator=data_collator,
            )
            result = trainer.evaluate()
            return EvalResult(loss=result["eval_loss"])

    def __preprocess_function(self, examples):
        """
        :param examples:
        :return:
        """
        model_inputs = self.tokenizer(examples["input"], max_length=self.__max_input_length, truncation=True)

        # Setup the tokenizer for targets
        with self.tokenizer.as_target_tokenizer():
            labels = self.tokenizer(examples["target"], max_length=self.__max_output_length, truncation=True)

        model_inputs["labels"] = labels["input_ids"]
        return model_inputs



    def test(self, input_filepath, solve, args=TTTestArgs):
        raise NotImplementedError()



from pytest import approx

from happytransformer import (
    HappyWordPrediction,
    ARGS_WP_TRAIN,
    ARGS_WP_EVAl,
    WPTrainArgs,
    WPEvalArgs
)
from happytransformer.happy_word_prediction import WordPredictionResult
from tests.shared_tests import run_save_load


def test_wp_basic():
    MODELS = [
        ('DISTILBERT', 'distilbert-base-uncased', 'pepper'),
        ('BERT', 'bert-base-uncased', '.'),
        ('ALBERT', 'albert-base-v2', 'garlic'),
        ('ROBERTA', "roberta-base", ' pepper')  # todo look into why roberta predicts a space
    ]
    for model_type, model_name, top_result in MODELS:
        happy_wp = HappyWordPrediction(model_type, model_name)
        results = happy_wp.predict_mask(
            "Please pass the salt and [MASK]",
        )
        result = results[0]
        assert result.token == top_result


def test_wp_high_k():

    happy_wp = HappyWordPrediction("ALBERT", "albert-base-v2")
    results = happy_wp.predict_mask(
        "Please pass the salt and [MASK]", top_k=3000
    )
    assert results[0].token == "garlic"

def test_wp_top_k():
    happy_wp = HappyWordPrediction('DISTILBERT', 'distilbert-base-uncased')
    result = happy_wp.predict_mask(
        "Please pass the salt and [MASK]",
        top_k=2
    )
    answer = [
        WordPredictionResult(token='pepper', score=approx(0.2664579749107361, 0.01)),
        WordPredictionResult(token='vinegar', score=approx(0.08760260790586472, 0.01))
    ]

    assert result == answer


def test_wp_targets():
    happy_wp = HappyWordPrediction('DISTILBERT', 'distilbert-base-uncased')
    result = happy_wp.predict_mask(
        "Please pass the salt and [MASK]",
        targets=["water", "spices"], top_k=2
    )
    assert result[0].token == "water" and result[1].token == "spices"
    assert type(result[0].score) == float
    assert 0.0138 <= result[0].score <= 0.0158


def test_wp_train_default():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    happy_wp.train("../data/wp/train-eval.txt")

def test_wp_train_line_by_line():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    happy_wp.train("../data/wp/train-eval.txt", args=WPTrainArgs(line_by_line=True))



def test_wp_eval_basic():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    result = happy_wp.eval("../data/wp/train-eval.txt")
    assert type(result.loss) == float

def test_wp_train_effectiveness_multi():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')

    before_result = happy_wp.eval("../data/wp/train-eval.txt")

    happy_wp.train("../data/wp/train-eval.txt")
    after_result = happy_wp.eval("../data/wp/train-eval.txt")

    assert after_result.loss < before_result.loss

def test_wp_eval_some_settings():
    """
    Test to see what happens when only a subset of the potential settings are used
    :return:
    """
    args = {'line_by_line': True,
            }
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    result = happy_wp.eval("../data/wp/train-eval.txt", args)
    assert type(result.loss) == float


def test_wp_save_load_train():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    output_path = "data/wp-train.json"
    data_path = "../data/wp/train-eval.txt"
    args = ARGS_WP_TRAIN
    args["line_by_line"] = True
    run_save_load(happy_wp, output_path, args, data_path, "train")

def test_wp_save_load_eval():
    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    output_path = "data/wp-eval.json"
    data_path = "../data/wp/train-eval.txt"
    args = ARGS_WP_EVAl
    args["line_by_line"] = True
    run_save_load(happy_wp, output_path, args, data_path, "eval")

def test_wp_save():
    happy = HappyWordPrediction("BERT", "prajjwal1/bert-tiny")
    happy.save("model/")
    result_before = happy.predict_mask("I think therefore I [MASK]")

    happy = HappyWordPrediction(load_path="model/")
    result_after = happy.predict_mask("I think therefore I [MASK]")

    assert result_before[0].token ==result_after[0].token


def test_wp_train_eval_with_dic():

    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    train_args = {'learning_rate': 0.01, 'line_by_line': True, "num_train_epochs": 1}


    happy_wp.train("../data/wp/train-eval.txt" , args=train_args)
    eval_args = {'line_by_line': True}

    after_result = happy_wp.eval("../data/wp/train-eval.txt", args=eval_args)


def test_wp_train_eval_with_dataclass():

    happy_wp = HappyWordPrediction('', 'distilroberta-base')
    train_args = WPTrainArgs(learning_rate=0.01, line_by_line=True, num_train_epochs=1)

    happy_wp.train("../data/wp/train-eval.txt" , args=train_args)

    eval_args = WPEvalArgs(line_by_line=True)

    after_result = happy_wp.eval("../data/wp/train-eval.txt", args=eval_args)




from .adaptor import Adaptor



class RobertaAdaptor(Adaptor):

    @staticmethod
    def preprocess_mask_text(text: str) -> str:
        return text.replace('[MASK]', '<mask>')

    @staticmethod
    def postprocess_mask_prediction_token(text):
        if text:
            return text[1:] if text[0] == "Ġ" else text
        else:
            return ""

class AlbertAdaptor(Adaptor):
    @staticmethod
    def postprocess_mask_prediction_token(text):
        if text:
            return text[1:] if text[0] == "▁" else text
        else:
            return ""



import torch
from transformers import AutoModelForNextSentencePrediction
from happytransformer.happy_transformer import HappyTransformer
from happytransformer.adaptors import get_adaptor
from happytransformer.qa import ARGS_QA_TRAIN, ARGS_QA_EVAl, ARGS_QA_TEST

class HappyNextSentence(HappyTransformer):
    """
    A user facing class for next sentence prediction
    """
    def __init__(self, model_type="BERT",
                 model_name="bert-base-uncased", 
                 load_path: str = "", 
                 use_auth_token: str = None):

        self.adaptor = get_adaptor(model_type)
        if load_path != "":
            model = AutoModelForNextSentencePrediction.from_pretrained(load_path)
        else:
            model = AutoModelForNextSentencePrediction.from_pretrained(model_name, use_auth_token=use_auth_token)

        super().__init__(model_type, model_name, model)
        self._pipeline = None
        self._trainer = None

    def predict_next_sentence(self, sentence_a: str, sentence_b: str) -> float:
        """
        Predict the probability that sentence_b follows sentence_a.
        Higher probabilities indicate more coherent sentence pairs.
        """

        encoded = self.tokenizer(sentence_a, sentence_b, return_tensors='pt')
        with torch.no_grad():
            scores = self.model(encoded['input_ids'], token_type_ids=encoded['token_type_ids']).logits[0]

        probabilities = torch.softmax(scores, dim=0)
        # probability that sentence B follows sentence A
        score = probabilities[0].item()

        if self._device == 'cuda':
            torch.cuda.empty_cache()

        return score

    def train(self, input_filepath, args=ARGS_QA_TRAIN):
        raise NotImplementedError("train() is currently not available")

    def eval(self, input_filepath, args=ARGS_QA_EVAl):
        raise NotImplementedError("eval() is currently not available")

    def test(self, input_filepath, args=ARGS_QA_TEST):
        raise NotImplementedError("test() is currently not available")



"""
Contains a class called HappyTextToText which performs text to text generation
"""
from dataclasses import dataclass

from transformers import Text2TextGenerationPipeline, AutoModelForSeq2SeqLM

from happytransformer.happy_transformer import HappyTransformer
from happytransformer.tt.trainer import TTTrainer
from happytransformer.cuda_detect import detect_cuda_device_number
from happytransformer.adaptors import get_adaptor
from happytransformer.tt.trainer import TTTrainArgs, TTEvalArgs, TTTestArgs


@dataclass
class TextToTextResult:
    """
    Returned when HappyTextToText.generate() is called
    """
    text: str

@dataclass
class TTSettings:
    """
    Used to adjust the text generation algorithm that's used when
    HappyTextToText.generate() is called 

    """
    min_length: int = 10
    max_length: int = 50
    do_sample: bool = False
    early_stopping: bool = False
    num_beams: int = 1
    temperature: float = 1
    top_k: int = 50
    no_repeat_ngram_size: int = 0
    top_p: float = 1


class HappyTextToText(HappyTransformer):
    """
    A user facing class for text to text generation
    """
    def __init__(self, model_type: str = "T5", model_name: str = "t5-small", load_path: str = "", use_auth_token: str = None):

        self.adaptor = get_adaptor(model_type)

        if load_path != "":
            model = AutoModelForSeq2SeqLM.from_pretrained(load_path)
        else:
            model = AutoModelForSeq2SeqLM.from_pretrained(model_name, use_auth_token=use_auth_token)


        super().__init__(model_type, model_name, model)

        device_number = detect_cuda_device_number()

        self._pipeline = Text2TextGenerationPipeline(model=self.model,
                                                     tokenizer=self.tokenizer, device=device_number)

        self._trainer = TTTrainer(self.model, model_type, self.tokenizer, self._device, self.logger)


    def __assert_default_text_is_val(self, text):
        """
        Ensures the input's text input is valid.
        Raises a Value Error if the text input is not valid.
        :param text: The value the user inputs for the "text" parameter
        """

        if not isinstance(text, str):
            raise ValueError("The text input must be a string")
        if not text:
            raise ValueError("The text input must have at least one character")


    def generate_text(self, text: str,
                      args: TTSettings = TTSettings()) -> TextToTextResult:
        """
        :param text: starting text that the model uses as a prompt to continue it.
        :param args: A TTSettings object
        :return: A TextToTextResult() object
        """
        self.__assert_default_text_is_val(text)

        output = self._pipeline(text, min_length=args.min_length,
                                max_length=args.max_length,
                                do_sample=args.do_sample,
                                early_stopping=args.early_stopping,
                                num_beams=args.num_beams,
                                temperature=args.temperature,
                                top_k=args.top_k,
                                no_repeat_ngram_size=args.no_repeat_ngram_size,
                                top_p=args.top_p,
                                )
        return TextToTextResult(text=output[0]['generated_text'])

    def train(self, input_filepath, args=TTTrainArgs()):
        """
        Trains the text-to-text model
        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values: text_1, text_2
        args: A TTTrainArgs() object
        return: None
        """
        self._trainer.train(input_filepath=input_filepath, dataclass_args=args)

    def eval(self, input_filepath, args=TTEvalArgs()):
        """
        Evaluated the text-to-text model

        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values: text_1, text_2

        args: A TTEvalArgs() object
        return: an EvalResult() object
        """

        result = self._trainer.eval(input_filepath=input_filepath, dataclass_args=args)
        return result


    def test(self, input_filepath, args=TTTestArgs):
        raise NotImplementedError("test() is currently not available")



"""
Contains the HappyGeneration class
"""
from dataclasses import dataclass
from typing import List
from transformers import AutoModelForCausalLM, TextGenerationPipeline
from happytransformer.happy_transformer import HappyTransformer
from happytransformer.gen.trainer import GENTrainer, GENTrainArgs, GENEvalArgs
from happytransformer.adaptors import get_adaptor
from happytransformer.gen import ARGS_GEN_TRAIN, ARGS_GEN_EVAl, ARGS_GEN_TEST
from happytransformer.happy_trainer import EvalResult
from happytransformer.fine_tuning_util import create_args_dataclass
from happytransformer.cuda_detect import detect_cuda_device_number

"""
The main settings that users will adjust when performing experiments

The values for full_settings are the same as the default values above except for min and max length. 
"""
@dataclass
class GENSettings:
    min_length: int = 10
    max_length: int = 50
    do_sample: bool = False
    early_stopping: bool = False
    num_beams: int = 1
    temperature: float = 1
    top_k: int = 50
    no_repeat_ngram_size: int = 0
    top_p: float = 1
    bad_words: List[str] = None

@dataclass
class GenerationResult:
    text: str


class HappyGeneration(HappyTransformer):
    """
    This class is a user facing class that allows users to generate text using
    text generation Transformer models.

    The purpose of this class is to be lightweight and easy
    to understand and to offload complex tasks to
    other classes.
    """
    def __init__(self, model_type: str = "GPT2", model_name: str = "gpt2", 
                 load_path: str = "", use_auth_token: str = None):

        self.adaptor = get_adaptor(model_type)

        if load_path != "":
            model = AutoModelForCausalLM.from_pretrained(load_path)
        else:
            model = AutoModelForCausalLM.from_pretrained(model_name, use_auth_token=use_auth_token)

        super().__init__(model_type, model_name, model)
        device_number = detect_cuda_device_number()

        self._pipeline = TextGenerationPipeline(model=self.model, tokenizer=self.tokenizer, device=device_number)

        self._trainer = GENTrainer(self.model, model_type, self.tokenizer, self._device, self.logger)

    def __assert_default_text_is_val(self, text):
        """
        Ensures the input's text input is valid.
        Raises a Value Error if the text input is not valid.
        :param text: The value the user inputs for the "text" parameter
        """

        if not isinstance(text, str):
            raise ValueError("The text input must be a string")
        if not text:
            raise ValueError("The text input must have at least one character")


    def generate_text(self, text: str, args: GENSettings=GENSettings()) -> GenerationResult:
        """
        :param text: starting text that the model uses as a prompt to continue it.
        :param args: A GENSettings object
        :return: A GenerationResult() object
        """

        self.__assert_default_text_is_val(text)
        input_ids = self.tokenizer.encode(text, return_tensors="pt")
        adjusted_min_length = args.min_length + len(input_ids[0])
        adjusted_max_length = args.max_length + len(input_ids[0])
        if args.bad_words:
            bad_words_ids = [self.tokenizer(" "+phrase.strip()).input_ids for phrase in args.bad_words]
        else:
            bad_words_ids = None
        output = self._pipeline(text, min_length=adjusted_min_length,
                                return_full_text=False,
                                max_length=adjusted_max_length,
                                do_sample=args.do_sample,
                                early_stopping=args.early_stopping,
                                num_beams=args.num_beams,
                                temperature=args.temperature,
                                top_k=args.top_k,
                                no_repeat_ngram_size=args.no_repeat_ngram_size,
                                top_p=args.top_p,
                                bad_words_ids=bad_words_ids
                                )
        return GenerationResult(text=output[0]['generated_text'])


    def __post_process_generated_text(self, result, text):
        """
        A method for processing the output of the model. More features will be added later.
        :param result: result the output of the model after being decoded
        :param text:  the original input to generate_text
        :return: returns to text after going through post-processing. Removes starting text
        """

        return result[len(text):]


    def train(self, input_filepath: str, args=GENTrainArgs()):
        """
        :param input_filepath:a file path to a text file that contains nothing but training data
        :param args: either a GENTrainArgs() object or a dictionary that contains all of the same keys as ARGS_GEN_TRAIN
        :return: None
        """

        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_GEN_TRAIN,
                                                         input_dic_args=args,
                                                         method_dataclass_args=GENTrainArgs)
        elif type(args) == GENTrainArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a GENTrainArgs object or a dictionary")

        self._trainer.train(input_filepath=input_filepath, dataclass_args=method_dataclass_args)

    def eval(self, input_filepath: str, args=GENEvalArgs()) -> EvalResult:
        """
        :param input_filepath:a file path to a text file that contains nothing but evaluating data
        :param args: either a GENEvalArgs() object or a dictionary that contains all of the same keys as ARGS_GEN_EVAl
        :return: None
        """
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_GEN_EVAl,
                                                         input_dic_args=args,
                                                         method_dataclass_args=GENEvalArgs)
        elif type(args) == GENEvalArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a GENEvalArgs object or a dictionary")

        return self._trainer.eval(input_filepath=input_filepath, dataclass_args=method_dataclass_args)

    def test(self, input_filepath, args=ARGS_GEN_TEST):
        raise NotImplementedError("test() is currently not available")



"""
This code is a modified version of the official documentation for the
transformer library by Hugging Face which can be found below.

We prioritized following the official documentation as close as possible to ensure we're using
robust methods. And also, to improve maintainability as they update the documentation.

https://huggingface.co/transformers/custom_datasets.html#sequence-classification-with-imdb-reviews"""

from dataclasses import dataclass
import csv
import torch
from happytransformer.happy_trainer import HappyTrainer, EvalResult
from transformers import DataCollatorWithPadding
from tqdm import tqdm
from happytransformer.tc.default_args import ARGS_TC_TRAIN, ARGS_TC_EVAL, ARGS_TC_TEST
import json

@dataclass
class TCTrainArgs:
    learning_rate: float = ARGS_TC_TRAIN["learning_rate"]
    num_train_epochs: int = ARGS_TC_TRAIN["num_train_epochs"]
    batch_size: int = ARGS_TC_TRAIN["batch_size"]
    weight_decay: float = ARGS_TC_TRAIN["weight_decay"]
    adam_beta1: float = ARGS_TC_TRAIN["adam_beta1"]
    adam_beta2: float = ARGS_TC_TRAIN["adam_beta2"]
    adam_epsilon: float = ARGS_TC_TRAIN["adam_epsilon"]
    max_grad_norm:  float = ARGS_TC_TRAIN["max_grad_norm"]
    save_preprocessed_data: bool = ARGS_TC_TRAIN["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_TC_TRAIN["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_TC_TRAIN["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_TC_TRAIN["load_preprocessed_data_path"]


@dataclass
class TCEvalArgs:
    save_preprocessed_data: bool = ARGS_TC_EVAL["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_TC_EVAL["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_TC_EVAL["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_TC_EVAL["load_preprocessed_data_path"]
    batch_size: int = ARGS_TC_EVAL["batch_size"]


@dataclass
class TCTestArgs:
    save_preprocessed_data: bool = ARGS_TC_TEST["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_TC_TEST["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_TC_TEST["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_TC_TEST["load_preprocessed_data_path"]

class TCTrainer(HappyTrainer):
    """
    A class for training text classification functionality
    """

    def train(self, input_filepath, dataclass_args: TCTrainArgs):

        if not dataclass_args.load_preprocessed_data:
            self.logger.info("Preprocessing dataset...")
            contexts, labels = self._get_data(input_filepath)
            train_encodings = self.tokenizer(contexts, truncation=True, padding=True)
        else:
            self.logger.info("Loading dataset from %s...", dataclass_args.load_preprocessed_data_path)
            train_encodings, labels = self._get_preprocessed_data(dataclass_args.load_preprocessed_data_path)

        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving training dataset to %s...", dataclass_args.save_preprocessed_data_path)
            input_ids = train_encodings["input_ids"]
            attention_mask = train_encodings["attention_mask"]
            self._generate_json(dataclass_args.save_preprocessed_data_path, input_ids, attention_mask, labels, "train")

        train_dataset = TextClassificationDataset(train_encodings, labels)
        data_collator = DataCollatorWithPadding(self.tokenizer)
        self._run_train(train_dataset, dataclass_args, data_collator)

    def eval(self, input_filepath, dataclass_args: TCEvalArgs):
        if not dataclass_args.load_preprocessed_data:
            self.logger.info("Preprocessing dataset...")
            contexts, labels = self._get_data(input_filepath)
            eval_encodings = self.tokenizer(contexts, truncation=True, padding=True)
        else:
            self.logger.info("Loading dataset from %s...", dataclass_args.load_preprocessed_data_path)
            eval_encodings, labels = self._get_preprocessed_data(dataclass_args.load_preprocessed_data_path)

        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving training dataset to %s...", dataclass_args.save_preprocessed_data_path)
            input_ids = eval_encodings["input_ids"]
            attention_mask = eval_encodings["attention_mask"]
            self._generate_json(dataclass_args.save_preprocessed_data_path, input_ids, attention_mask, labels, "train")


        eval_dataset = TextClassificationDataset(eval_encodings, labels)
        data_collator = DataCollatorWithPadding(self.tokenizer)

        result = self._run_eval(eval_dataset, data_collator, dataclass_args)
        return EvalResult(loss=result["eval_loss"])

    def test(self, input_filepath, solve, dataclass_args: TCTestArgs):
        """
        See docstring in HappyQuestionAnswering.test()
        solve: HappyQuestionAnswering.answers_to_question()
        """
        contexts = self._get_data(input_filepath, test_data=True)

        return [
            solve(context)
            for context in tqdm(contexts)
        ]

    @staticmethod
    def _get_data(filepath, test_data=False):
        """
        Used for parsing data for training and evaluating (both contain labels)
        :param filepath: a string that contains the location of the data
        :return:
        """
        contexts = []
        labels = []
        with open(filepath, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            for row in reader:
                contexts.append(row['text'])
                if not test_data:
                    labels.append(int(row['label']))
        csv_file.close()

        if not test_data:
            return contexts, labels
        return contexts

    @staticmethod
    def _generate_json(json_path, input_ids, attention_mask, labels, name):
        data = {}
        data[name] = []
        data = {
            name: [
                {
                    'attention_mask': input_id,
                    'input_ids': attention_mask,
                    'labels': label
                }
                for input_id, attention_mask, label in zip(input_ids, attention_mask, labels)
            ]
        }

        with open(json_path, 'w') as outfile:
            json.dump(data, outfile)

    @staticmethod
    def _get_preprocessed_data(filepath):
        """
        Used for Fetching preprocessed data)
        :param filepath: a string that contains the location of the data
        :return:
        """

        # dataset = load_dataset("csv", data_files={"train": filepath})
        input_ids = []
        attention_mask = []
        labels = []
        with open(filepath) as json_file:
            data = json.load(json_file)
        json_file.close()

        for case in data["train"]:
            input_ids.append(case['input_ids'])
            attention_mask.append(case['attention_mask'])
            labels.append(case['labels'])


        train_encodings = {
            "input_ids": input_ids,
            "attention_mask": attention_mask
        }

        return train_encodings, labels


class TextClassificationDataset(torch.utils.data.Dataset):
    """
    A class to allow the training and testing data to be used by
    a transformers.Trainer object
    """
    def __init__(self, encodings, labels):
        self.encodings = encodings
        self.labels = labels

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        item['labels'] = torch.tensor(self.labels[idx])
        return item

    def __len__(self):
        return len(self.labels)

class TextClassificationDatasetTest(torch.utils.data.Dataset):
    """
    A class to allow the testing data to be used by
    a transformers.Trainer object
    """
    def __init__(self, encodings, length):
        self.encodings = encodings
        self.length = length

    def __getitem__(self, idx):
        item = {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}
        return item

    def __len__(self):
        return self.length



"""
Contains a class called HappyTextClassification that performs text classification
"""
from dataclasses import dataclass

from transformers import TextClassificationPipeline, AutoConfig, AutoModelForSequenceClassification

from happytransformer.tc.trainer import TCTrainer, TCTrainArgs, TCEvalArgs, TCTestArgs
from happytransformer.cuda_detect import detect_cuda_device_number
from happytransformer.happy_transformer import HappyTransformer
from happytransformer.adaptors import get_adaptor
from happytransformer.tc import ARGS_TC_TRAIN, ARGS_TC_EVAL, ARGS_TC_TEST
from happytransformer.happy_trainer import EvalResult
from happytransformer.fine_tuning_util import create_args_dataclass


@dataclass
class TextClassificationResult:
    label: str
    score: float

class HappyTextClassification(HappyTransformer):
    """
    A user facing class for Text Classification
    """

    def __init__(self, model_type="DISTILBERT",
                 model_name="distilbert-base-uncased", num_labels: int = 2, load_path: str = "", use_auth_token: str = None):
        self.adaptor = get_adaptor(model_type)

        config = AutoConfig.from_pretrained(model_name, num_labels=num_labels)

        if load_path != "":
            model = AutoModelForSequenceClassification.from_pretrained(load_path, config=config)
        else:
            model = AutoModelForSequenceClassification.from_pretrained(model_name, config=config, use_auth_token=use_auth_token)


        super().__init__(model_type, model_name, model)

        device_number = detect_cuda_device_number()
        self._pipeline = TextClassificationPipeline(
            model=self.model, tokenizer=self.tokenizer,
            device=device_number
        )

        self._trainer = TCTrainer(
            self.model, self.model_type,
            self.tokenizer, self._device, self.logger
        )

    def classify_text(self, text: str) -> TextClassificationResult:
        """
        Classify text to a label based on model's training
        """
        # Blocking allowing a for a list of strings
        if not isinstance(text, str):
            raise ValueError("the \"text\" argument must be a single string")
        results = self._pipeline(text)
        # we do not support predicting a list of  texts, so only first prediction is relevant
        first_result = results[0]

        return TextClassificationResult(label=first_result["label"], score=first_result["score"])
    
    def train(self, input_filepath, args=TCTrainArgs()):
        """
        Trains the question answering model
        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values: text, label
        args: Either a TCTrainArgs() object or a dictionary that contains all of the same keys as ARGS_TC_TRAIN
        return: None
        """
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_TC_TRAIN,
                                                         input_dic_args=args,
                                                         method_dataclass_args=TCTrainArgs)
        elif type(args) == TCTrainArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a TCTrainArgs object or a dictionary")

        self._trainer.train(input_filepath=input_filepath, dataclass_args=method_dataclass_args)

    def eval(self, input_filepath, args=TCEvalArgs()) -> EvalResult:
        """
        Evaluated the text classification answering model
        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header values:
        text, label

        return: an EvalResult() object
        """
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_TC_EVAL,
                                                          input_dic_args=args,
                                                          method_dataclass_args=TCEvalArgs)
        elif type(args) == TCEvalArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a TCEvalArgs object or a dictionary")

        return self._trainer.eval(input_filepath=input_filepath, dataclass_args=method_dataclass_args)


    def test(self, input_filepath, args=TCTestArgs()):
        """
        Tests the text classification  model. Used to obtain results
        input_filepath: a string that contains the location of a csv file
        for training. Contains the following header value:
         text
        return: A list of TextClassificationResult() objects
        """

        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_TC_TEST,
                                                          input_dic_args=args,
                                                          method_dataclass_args=TCTestArgs)
        elif type(args) == TCTestArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a TCTestArgs() object or a dictionary")

        return self._trainer.test(input_filepath=input_filepath, solve=self.classify_text, dataclass_args=method_dataclass_args)



"""

Fine-tuning for text generation models.

Based on the tutorial found here:
https://github.com/huggingface/transformers/blob/master/examples/pytorch/language-modeling/run_clm.py
"""
from dataclasses import dataclass
import json
from transformers import default_data_collator
from happytransformer.happy_trainer import HappyTrainer, EvalResult
from happytransformer.fine_tuning_util import preprocess_concatenate
from happytransformer.gen.default_args import ARGS_GEN_TRAIN, ARGS_GEN_EVAl
from datasets import load_dataset

@dataclass
class GENTrainArgs:
    learning_rate: float = ARGS_GEN_TRAIN["learning_rate"]
    num_train_epochs: int = ARGS_GEN_TRAIN["num_train_epochs"]
    batch_size: int = ARGS_GEN_TRAIN["batch_size"]
    weight_decay: float = ARGS_GEN_TRAIN["weight_decay"]
    adam_beta1: float = ARGS_GEN_TRAIN["adam_beta1"]
    adam_beta2: float = ARGS_GEN_TRAIN["adam_beta2"]
    adam_epsilon: float = ARGS_GEN_TRAIN["adam_epsilon"]
    max_grad_norm:  float = ARGS_GEN_TRAIN["max_grad_norm"]
    save_preprocessed_data: bool = ARGS_GEN_TRAIN["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_GEN_TRAIN["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_GEN_TRAIN["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_GEN_TRAIN["load_preprocessed_data_path"]
    preprocessing_processes: int = ARGS_GEN_TRAIN["preprocessing_processes"]
    mlm_probability: float = ARGS_GEN_TRAIN["mlm_probability"]


@dataclass
class GENEvalArgs:
    batch_size: int = ARGS_GEN_EVAl["batch_size"]
    save_preprocessed_data: bool = ARGS_GEN_EVAl["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_GEN_EVAl["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_GEN_EVAl["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_GEN_EVAl["load_preprocessed_data_path"]
    preprocessing_processes: int =ARGS_GEN_EVAl["preprocessing_processes"]
    mlm_probability: float = ARGS_GEN_EVAl["mlm_probability"]


class GENTrainer(HappyTrainer):
    """
    Trainer class for HappyWordPrediction
    """

    def train(self, input_filepath, dataclass_args: GENTrainArgs):
        """
        :param input_filepath: A file path to a text file that contains nothing but training data
        :param dataclass_args: A GENTrainArgs() object
        :return: None
        """

        if not dataclass_args.load_preprocessed_data:
            self.logger.info("Preprocessing dataset...")
            dataset = load_dataset("text", data_files={"train": input_filepath})
            tokenized_dataset = preprocess_concatenate(self.tokenizer, dataset, dataclass_args.preprocessing_processes, False)

        else:
            self.logger.info("Loading dataset from %s...", dataclass_args.load_preprocessed_data_path)
            tokenized_dataset = load_dataset("json", data_files={"train": dataclass_args.load_preprocessed_data_path}, field='train')

        if dataclass_args.save_preprocessed_data:
            if dataclass_args.load_preprocessed_data:
                self.logger.warning("Both save_preprocessed_data and load_data are enabled,")

            self.logger.info("Saving training dataset to %s...", dataclass_args.save_preprocessed_data_path)

            self._generate_json(dataclass_args.save_preprocessed_data_path, tokenized_dataset["train"], "train")

        self.logger.info("Training...")

        self._run_train(tokenized_dataset['train'], dataclass_args, default_data_collator)

    def eval(self, input_filepath, dataclass_args: GENEvalArgs):
        """
        :param input_filepath: A file path to a text file that contains nothing but evaluating data
        :param dataclass_args: A GENEvalArgs() object
        :return: An EvalResult() object
        """

        if not dataclass_args.load_preprocessed_data:
            self.logger.info("Preprocessing dataset...")
            datasets = load_dataset("text", data_files={"eval": input_filepath})
            tokenized_dataset = preprocess_concatenate(self.tokenizer, datasets, dataclass_args.preprocessing_processes, False)

        else:
            self.logger.info("Loading dataset from %s...", dataclass_args.load_preprocessed_data_path)
            tokenized_dataset = load_dataset("json", data_files={"eval": dataclass_args.load_preprocessed_data_path}, field='eval')

        if dataclass_args.save_preprocessed_data:
            if dataclass_args. load_preprocessed_data:
                self.logger.warning("Both save_preprocessed_data and load_data are enabled.")

            self.logger.info("Saving evaluating dataset to %s...", dataclass_args.save_preprocessed_data_path)

            self._generate_json(dataclass_args.save_preprocessed_data_path, tokenized_dataset["eval"], "eval")

        self.logger.info("Evaluating...")

        result = self._run_eval(tokenized_dataset['eval'], default_data_collator, dataclass_args)
        return EvalResult(loss=result["eval_loss"])

    def test(self, input_filepath, solve, args):
        raise NotImplementedError()

    def _generate_json(self, json_path, dataset, name):
        """
        :param json_path: A path to a json file that will be created/overwritten
        :param dataset: A list of dictionaries that contain the keys "attention_mask," "input_ids" and "labels"
        :param name: A string to specify if the written data is for "Train" or "Eval"
        :return: None
        """
        data = {}
        data[name] = []
        data = {
            name: [
                {
                    'attention_mask': case['attention_mask'],
                    'input_ids': case['input_ids'],
                    'labels': case['labels']
                }
                for case in dataset
            ]
        }

        with open(json_path, 'w') as outfile:
            json.dump(data, outfile)



"""
This code is a modified version of the official documentation for the
transformer library by Hugging Face which can be found below.

We prioritized following the official documentation as close as possible to ensure we're using
robust methods. And also, to improve maintainability as they update the documentation.

https://huggingface.co/transformers/custom_datasets.html#question-answering-with-squad-2-0
"""

from dataclasses import dataclass
import csv
from tqdm import tqdm
import torch
import json
from transformers import DataCollatorWithPadding
from happytransformer.happy_trainer import HappyTrainer, EvalResult
from happytransformer.qa.default_args import ARGS_QA_TRAIN, ARGS_QA_EVAl, ARGS_QA_TEST

@dataclass
class QATrainArgs:
    learning_rate: float = ARGS_QA_TRAIN["learning_rate"]
    num_train_epochs: int = ARGS_QA_TRAIN["num_train_epochs"]
    weight_decay: float = ARGS_QA_TRAIN["weight_decay"]
    adam_beta1: float = ARGS_QA_TRAIN["adam_beta1"]
    adam_beta2: float = ARGS_QA_TRAIN["adam_beta2"]
    adam_epsilon: float = ARGS_QA_TRAIN["adam_epsilon"]
    max_grad_norm:  float = ARGS_QA_TRAIN["max_grad_norm"]
    save_preprocessed_data: bool = ARGS_QA_TRAIN["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_QA_TRAIN["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_QA_TRAIN["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_QA_TRAIN["load_preprocessed_data_path"]
    batch_size: int = ARGS_QA_TRAIN["batch_size"]


@dataclass
class QAEvalArgs:
    batch_size: int = ARGS_QA_EVAl["batch_size"]
    save_preprocessed_data: bool = ARGS_QA_EVAl["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_QA_EVAl["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_QA_EVAl["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_QA_EVAl["load_preprocessed_data_path"]


@dataclass
class QATestArgs:
    save_preprocessed_data: bool = ARGS_QA_TEST["save_preprocessed_data"]
    save_preprocessed_data_path: str = ARGS_QA_TEST["save_preprocessed_data_path"]
    load_preprocessed_data: bool = ARGS_QA_TEST["load_preprocessed_data"]
    load_preprocessed_data_path: str = ARGS_QA_TEST["load_preprocessed_data_path"]


class QATrainer(HappyTrainer):
    """
    Trainer class for HappyTextClassification
    """

    def train(self, input_filepath, dataclass_args: QATrainArgs):
        """
        See docstring in HappyQuestionAnswering.train()
        """
        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")
        if dataclass_args.load_preprocessed_data:
            self.logger.info("Loading preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")

        self.logger.info("Preprocessing dataset...")
        contexts, questions, answers = self._get_data(input_filepath)
        self.__add_end_idx(contexts, answers)
        encodings = self.tokenizer(contexts, questions, truncation=True, padding=True)
        self.__add_token_positions(encodings, answers)
        dataset = QuestionAnsweringDataset(encodings)
        data_collator = DataCollatorWithPadding(self.tokenizer)
        self._run_train(dataset, dataclass_args, data_collator)

    def eval(self, input_filepath, dataclass_args: QAEvalArgs):
        """
        See docstring in HappyQuestionAnswering.eval()

        """
        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")
        if dataclass_args.load_preprocessed_data:
            self.logger.info("Loading preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")

        contexts, questions, answers = self._get_data(input_filepath)

        self.__add_end_idx(contexts, answers)
        encodings = self.tokenizer(contexts, questions, truncation=True, padding=True)
        self.__add_token_positions(encodings, answers)
        eval_dataset = QuestionAnsweringDataset(encodings)
        data_collator = DataCollatorWithPadding(self.tokenizer)

        result = self._run_eval(eval_dataset, data_collator, dataclass_args)
        return EvalResult(loss=result["eval_loss"])


    def test(self, input_filepath, solve, dataclass_args: QATestArgs):
        """
        See docstring in HappyQuestionAnswering.test()

        """

        if dataclass_args.save_preprocessed_data:
            self.logger.info("Saving preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")
        if dataclass_args.load_preprocessed_data:
            self.logger.info("Loading preprocessed data is currently "
                             "not available for question answering models. "
                             "It will be added soon. ")

        contexts, questions = self._get_data(input_filepath, test_data=True)

        return [
            solve(context, question)[0]
            for context, question in
            tqdm(zip(contexts, questions))
        ]

    @staticmethod
    def _get_data(filepath, test_data=False):
        """
        Used to collect
        :param filepath: a string that contains the location of the data
        :return: if test_data = False contexts, questions, answers (all strings)
        else: contexts, questions
        """
        contexts = []
        questions = []
        answers = []
        with open(filepath, newline='') as csv_file:
            reader = csv.DictReader(csv_file)
            for row in reader:
                contexts.append(row['context'])
                questions.append(row['question'])
                if not test_data:
                    answer = {}
                    answer["answer_text"] = row['answer_text']
                    answer["answer_start"] = int(row['answer_start'])
                    answers.append(answer)
        csv_file.close()

        if not test_data:
            return contexts, questions, answers
        return contexts, questions

    @staticmethod
    def __add_end_idx(contexts, answers):
        for answer, context in zip(answers, contexts):

            gold_text = answer['answer_text']
            start_idx = answer['answer_start']
            end_idx = start_idx + len(gold_text)

            # todo (maybe): strip answer['text'] (remove white space from start and end)
            # sometimes squad answers are off by a character or two – fix this
            if context[start_idx:end_idx] == gold_text:
                answer['answer_end'] = end_idx
            elif context[start_idx - 1:end_idx - 1] == gold_text:
                answer['answer_start'] = start_idx - 1
                answer['answer_end'] = end_idx - 1
            elif context[start_idx - 2:end_idx - 2] == gold_text:
                answer['answer_start'] = start_idx - 2
                answer['answer_end'] = end_idx - 2  
            else:
                print("error: implement skipping training answer")

    def __add_token_positions(self, encodings, answers):
        start_positions = []
        end_positions = []
        for i in range(len(answers)):
            start_positions.append(encodings.char_to_token(i, answers[i]['answer_start']))
            end_positions.append(encodings.char_to_token(i, answers[i]['answer_end'] - 1))
            if start_positions[-1] is None:
                start_positions[-1] = self.tokenizer.model_max_length
            if end_positions[-1] is None:
                end_positions[-1] = self.tokenizer.model_max_length
        encodings.update({'start_positions': start_positions, 'end_positions': end_positions})


    @staticmethod
    def _generate_json(json_path, input_ids, attention_masks, start_positions, end_positions, name):
        data = {}
        data[name] = []
        data = {
            name: [
                {
                    'input_ids': input_id,
                    'attention_mask': attention_mask,
                    'start_positions': start_positions,
                    'end_positions': end_positions

                }
                for input_id, attention_mask, start_position, end_position in zip(input_ids, attention_masks, start_positions, end_positions)
            ]
        }

        with open(json_path, 'w') as outfile:
            json.dump(data, outfile)


    @staticmethod
    def _get_preprocessed_data(filepath):
        """
        Used for Fetching preprocessed data)
        :param filepath: a string that contains the location of the data
        :return:
        """

        # dataset = load_dataset("csv", data_files={"train": filepath})
        input_ids = []
        attention_mask = []
        start_positions = []
        end_positions = []

        with open(filepath) as json_file:
            data = json.load(json_file)
        json_file.close()

        for case in data["train"]:
            input_ids.append(case['input_ids'])
            attention_mask.append(case['attention_mask'])
            start_positions.append(case['start_positions'])
            end_positions.append(case['end_positions'])

        encodings = {
            "input_ids": input_ids,
            "attention_mask": attention_mask,
            "start_positions": start_positions,
            "end_positions": end_positions}

        return encodings


class QuestionAnsweringDataset(torch.utils.data.Dataset):
    """
    A class used to iterate through the training data.
    It used to create  a torch DataLoader object, so that the training data can be
    iterated through in batches easily.
    """
    def __init__(self, encodings):
        self.encodings = encodings

    def __getitem__(self, idx):
        return {key: torch.tensor(val[idx]) for key, val in self.encodings.items()}

    def __len__(self):
        return len(self.encodings.input_ids)




class Adaptor:
    '''
    Holds a few functions for implementation details.
    Does NOT store any state.
    '''

    @staticmethod
    def preprocess_mask_text(text: str) -> str:
        return text

    @staticmethod
    def postprocess_mask_prediction_token(text: str) -> str:
        return text


from typing import List, Optional
from dataclasses import dataclass

from transformers import FillMaskPipeline, AutoModelForMaskedLM, PretrainedConfig

from happytransformer.happy_transformer import HappyTransformer
from happytransformer.wp.trainer import WPTrainer, WPTrainArgs, WPEvalArgs
from happytransformer.cuda_detect import detect_cuda_device_number
from happytransformer.adaptors import get_adaptor
from happytransformer.wp import ARGS_WP_TRAIN, ARGS_WP_EVAl, ARGS_WP_TEST
from happytransformer.happy_trainer import EvalResult
from happytransformer.fine_tuning_util import create_args_dataclass

@dataclass
class WordPredictionResult:
    token: str
    score: float

class HappyWordPrediction(HappyTransformer):
    """
    A user facing class for text classification
    """
    def __init__(
            self, model_type: str = "DISTILBERT", model_name: str = "distilbert-base-uncased",
            load_path: str ="", use_auth_token: str = None):


        self.adaptor = get_adaptor(model_type)

        if load_path != "":
            model = AutoModelForMaskedLM.from_pretrained(load_path)
        else:
            model = AutoModelForMaskedLM.from_pretrained(model_name, use_auth_token=use_auth_token)

        super().__init__(model_type, model_name, model, load_path=load_path)

        device_number = detect_cuda_device_number()

        self._pipeline = FillMaskPipeline(model=self.model, tokenizer=self.tokenizer, device=device_number)

        self._trainer = WPTrainer(self.model, model_type, self.tokenizer, self._device, self.logger)

    def predict_mask(self, text: str, targets: Optional[List[str]] = None, top_k: int = 1) -> List[WordPredictionResult]:
        """
        Predict [MASK] tokens in a string.
        targets limit possible guesses if supplied.
        top_k describes number of targets to return*
        *top_k does not apply if targets is supplied
        """
        if not isinstance(text, str):
            raise ValueError('the "text" argument must be a single string')

        text_for_pipeline = self.adaptor.preprocess_mask_text(text)
        answers = self._pipeline(
            text_for_pipeline, 
            targets=targets, top_k=top_k
        )

        fix_token = self.adaptor.postprocess_mask_prediction_token
        return [
            WordPredictionResult(
                token=fix_token(answer["token_str"]), 
                score=answer["score"]
            )
            for answer in answers
        ]

    def train(self, input_filepath, args=ARGS_WP_TRAIN):
        if type(args) == dict:
            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_WP_TRAIN,
                                                         input_dic_args=args,
                                                         method_dataclass_args=WPTrainArgs)
        elif type(args) == WPTrainArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a WPTrainArgs object or a dictionary")

        self._trainer.train(input_filepath=input_filepath, dataclass_args=method_dataclass_args)

    def eval(self, input_filepath, args=ARGS_WP_EVAl) -> EvalResult:
        if type(args) == dict:

            method_dataclass_args = create_args_dataclass(default_dic_args=ARGS_WP_EVAl,
                                                         input_dic_args=args,
                                                         method_dataclass_args=WPEvalArgs)
        elif type(args) == WPEvalArgs:
            method_dataclass_args = args
        else:
            raise ValueError("Invalid args type. Use a ARGS_WP_EVAl object or a dictionary")

        return self._trainer.eval(input_filepath=input_filepath, dataclass_args=method_dataclass_args)


    def test(self, input_filepath, args=ARGS_WP_TEST):
        raise NotImplementedError("test() is currently not available")
